def long_to_bytes(n, blocksize: int = ...): ...
def long_to_base64(data, size: int = ...): ...
def int_arr_to_long(arr): ...
def base64_to_long(data): ...
def calculate_at_hash(access_token, hash_alg): ...
def base64url_decode(input): ...
def base64url_encode(input): ...
def timedelta_total_seconds(delta): ...
def ensure_binary(s): ...
