from typing import Any

def sign(payload, key, headers: Any | None = ..., algorithm=...): ...
def verify(token, key, algorithms, verify: bool = ...): ...
def get_unverified_header(token): ...
def get_unverified_headers(token): ...
def get_unverified_claims(token): ...
